# encoding: utf-8

from __future__ import unicode_literals, absolute_import

import os
import sys
import locale
from itertools import chain
from six import iterkeys, iteritems
from six.moves.configparser import ConfigParser

from .autocomplete import SIMPLE as default_completion, ALL_MODES
from .keys import cli_key_dispatch as cli_key_dispatch


class Struct(object):
    """Simple class for instantiating objects we can add arbitrary attributes
    to and use for various arbitrary things."""


def getpreferredencoding():
    """Get the user's preferred encoding."""
    return locale.getpreferredencoding() or sys.getdefaultencoding()


def can_encode(c):
    try:
        c.encode(getpreferredencoding())
        return True
    except UnicodeEncodeError:
        return False


def supports_box_chars():
    """Check if the encoding supports Unicode box characters."""
    return all(map(can_encode, u'│─└┘┌┐'))


def get_config_home():
    """Returns the base directory for bpython's configuration files."""
    xdg_config_home = os.environ.get('XDG_CONFIG_HOME', '~/.config')
    return os.path.join(xdg_config_home, 'bpython')


def default_config_path():
    """Returns bpython's default configuration file path."""
    return os.path.join(get_config_home(), 'config')


def fill_config_with_default_values(config, default_values):
    for section in iterkeys(default_values):
        if not config.has_section(section):
            config.add_section(section)

        for (opt, val) in iteritems(default_values[section]):
            if not config.has_option(section, opt):
                config.set(section, opt, '%s' % (val, ))


def loadini(struct, configfile):
    """Loads .ini configuration file and stores its values in struct"""

    config_path = os.path.expanduser(configfile)

    config = ConfigParser()
    defaults = {
        'general': {
            'arg_spec': True,
            'auto_display_list': True,
            'autocomplete_mode': default_completion,
            'color_scheme': 'default',
            'complete_magic_methods': True,
            'dedent_after': 1,
            'default_autoreload': False,
            'editor': os.environ.get('VISUAL', os.environ.get('EDITOR', 'vi')),
            'flush_output': True,
            'highlight_show_source': True,
            'hist_duplicates': True,
            'hist_file': '~/.pythonhist',
            'hist_length': 100,
            'paste_time': 0.02,
            'pastebin_confirm': True,
            'pastebin_expiry': '1week',
            'pastebin_helper': '',
            'pastebin_removal_url': 'https://bpaste.net/remove/$removal_id',
            'pastebin_show_url': 'https://bpaste.net/show/$paste_id',
            'pastebin_url': 'https://bpaste.net/json/new',
            'save_append_py': False,
            'single_undo_time': 1.0,
            'syntax': True,
            'tab_length': 4,
            'unicode_box': True
        },
        'keyboard': {
            'backspace': 'C-h',
            'beginning_of_line': 'C-a',
            'clear_line': 'C-u',
            'clear_screen': 'C-l',
            'clear_word': 'C-w',
            'copy_clipboard': 'F10',
            'cut_to_buffer': 'C-k',
            'delete': 'C-d',
            'down_one_line': 'C-n',
            'edit_config': 'F3',
            'edit_current_block': 'C-x',
            'end_of_line': 'C-e',
            'exit': '',
            'external_editor': 'F7',
            'help': 'F1',
            'incremental_search': 'M-s',
            'last_output': 'F9',
            'left': 'C-b',
            'pastebin': 'F8',
            'reimport': 'F6',
            'reverse_incremental_search': 'M-r',
            'right': 'C-f',
            'save': 'C-s',
            'search': 'C-o',
            'show_source': 'F2',
            'suspend': 'C-z',
            'toggle_file_watch': 'F5',
            'transpose_chars': 'C-t',
            'undo': 'C-r',
            'up_one_line': 'C-p',
            'yank_from_buffer': 'C-y'
        },
        'cli': {
            'suggestion_width': 0.8,
            'trim_prompts': False,
        },
        'curtsies': {
            'list_above': False,
            'right_arrow_completion': True,
        }}

    default_keys_to_commands = dict((value, key) for (key, value)
                                    in iteritems(defaults['keyboard']))

    fill_config_with_default_values(config, defaults)
    try:
        if not config.read(config_path):
            # No config file. If the user has it in the old place then complain
            if os.path.isfile(os.path.expanduser('~/.bpython.ini')):
                sys.stderr.write("Error: It seems that you have a config file at "
                                 "~/.bpython.ini. Please move your config file to "
                                 "%s\n" % default_config_path())
                sys.exit(1)
    except UnicodeDecodeError as e:
        sys.stderr.write("Error: Unable to parse config file at '{}' due to an "
                         "encoding issue. Please make sure to fix the encoding "
                         "of the file or remove it and then try again.\n".format(config_path))
        sys.exit(1)

    def get_key_no_doublebind(command):
        default_commands_to_keys = defaults['keyboard']
        requested_key = config.get('keyboard', command)

        try:
            default_command = default_keys_to_commands[requested_key]

            if (default_commands_to_keys[default_command] ==
                    config.get('keyboard', default_command)):
                setattr(struct, '%s_key' % default_command, '')
        except KeyError:
            pass

        return requested_key

    struct.config_path = config_path

    struct.dedent_after = config.getint('general', 'dedent_after')
    struct.tab_length = config.getint('general', 'tab_length')
    struct.auto_display_list = config.getboolean('general',
                                                 'auto_display_list')
    struct.syntax = config.getboolean('general', 'syntax')
    struct.arg_spec = config.getboolean('general', 'arg_spec')
    struct.paste_time = config.getfloat('general', 'paste_time')
    struct.single_undo_time = config.getfloat('general', 'single_undo_time')
    struct.highlight_show_source = config.getboolean('general',
                                                     'highlight_show_source')
    struct.hist_file = config.get('general', 'hist_file')
    struct.editor = config.get('general', 'editor')
    struct.hist_length = config.getint('general', 'hist_length')
    struct.hist_duplicates = config.getboolean('general', 'hist_duplicates')
    struct.flush_output = config.getboolean('general', 'flush_output')
    struct.default_autoreload = config.getboolean(
        'general', 'default_autoreload')

    struct.pastebin_key = get_key_no_doublebind('pastebin')
    struct.copy_clipboard_key = get_key_no_doublebind('copy_clipboard')
    struct.save_key = get_key_no_doublebind('save')
    struct.search_key = get_key_no_doublebind('search')
    struct.show_source_key = get_key_no_doublebind('show_source')
    struct.suspend_key = get_key_no_doublebind('suspend')
    struct.toggle_file_watch_key = get_key_no_doublebind('toggle_file_watch')
    struct.undo_key = get_key_no_doublebind('undo')
    struct.reimport_key = get_key_no_doublebind('reimport')
    struct.reverse_incremental_search_key = get_key_no_doublebind(
        'reverse_incremental_search')
    struct.incremental_search_key = get_key_no_doublebind('incremental_search')
    struct.up_one_line_key = get_key_no_doublebind('up_one_line')
    struct.down_one_line_key = get_key_no_doublebind('down_one_line')
    struct.cut_to_buffer_key = get_key_no_doublebind('cut_to_buffer')
    struct.yank_from_buffer_key = get_key_no_doublebind('yank_from_buffer')
    struct.clear_word_key = get_key_no_doublebind('clear_word')
    struct.backspace_key = get_key_no_doublebind('backspace')
    struct.clear_line_key = get_key_no_doublebind('clear_line')
    struct.clear_screen_key = get_key_no_doublebind('clear_screen')
    struct.delete_key = get_key_no_doublebind('delete')

    struct.left_key = get_key_no_doublebind('left')
    struct.right_key = get_key_no_doublebind('right')
    struct.end_of_line_key = get_key_no_doublebind('end_of_line')
    struct.beginning_of_line_key = get_key_no_doublebind('beginning_of_line')
    struct.transpose_chars_key = get_key_no_doublebind('transpose_chars')
    struct.exit_key = get_key_no_doublebind('exit')
    struct.last_output_key = get_key_no_doublebind('last_output')
    struct.edit_config_key = get_key_no_doublebind('edit_config')
    struct.edit_current_block_key = get_key_no_doublebind('edit_current_block')
    struct.external_editor_key = get_key_no_doublebind('external_editor')
    struct.help_key = get_key_no_doublebind('help')

    struct.pastebin_confirm = config.getboolean('general', 'pastebin_confirm')
    struct.pastebin_url = config.get('general', 'pastebin_url')
    struct.pastebin_show_url = config.get('general', 'pastebin_show_url')
    struct.pastebin_removal_url = config.get('general', 'pastebin_removal_url')
    struct.pastebin_expiry = config.get('general', 'pastebin_expiry')
    struct.pastebin_helper = config.get('general', 'pastebin_helper')

    struct.cli_suggestion_width = config.getfloat('cli',
                                                  'suggestion_width')
    struct.cli_trim_prompts = config.getboolean('cli',
                                                'trim_prompts')

    struct.complete_magic_methods = config.getboolean('general',
                                                      'complete_magic_methods')
    struct.autocomplete_mode = config.get('general', 'autocomplete_mode')
    struct.save_append_py = config.getboolean('general', 'save_append_py')

    struct.curtsies_list_above = config.getboolean('curtsies', 'list_above')
    struct.curtsies_right_arrow_completion = \
        config.getboolean('curtsies', 'right_arrow_completion')

    color_scheme_name = config.get('general', 'color_scheme')

    default_colors = {
        'keyword': 'y',
        'name': 'c',
        'comment': 'b',
        'string': 'm',
        'error': 'r',
        'number': 'G',
        'operator': 'Y',
        'punctuation': 'y',
        'token': 'C',
        'background': 'd',
        'output': 'w',
        'main': 'c',
        'paren': 'R',
        'prompt': 'c',
        'prompt_more': 'g',
        'right_arrow_suggestion': 'K',
    }

    if color_scheme_name == 'default':
        struct.color_scheme = default_colors
    else:
        struct.color_scheme = dict()

        theme_filename = color_scheme_name + '.theme'
        path = os.path.expanduser(os.path.join(get_config_home(),
                                               theme_filename))
        try:
            load_theme(struct, path, struct.color_scheme, default_colors)
        except EnvironmentError:
            sys.stderr.write("Could not load theme '%s'.\n" %
                             (color_scheme_name, ))
            sys.exit(1)

    # expand path of history file
    struct.hist_file = os.path.expanduser(struct.hist_file)

    # verify completion mode
    if struct.autocomplete_mode not in ALL_MODES:
        struct.autocomplete_mode = default_completion

    # set box drawing characters
    if config.getboolean('general', 'unicode_box') and supports_box_chars():
        struct.left_border = '│'
        struct.right_border = '│'
        struct.top_border = '─'
        struct.bottom_border = '─'
        struct.left_bottom_corner = '└'
        struct.right_bottom_corner = '┘'
        struct.left_top_corner = '┌'
        struct.right_top_corner = '┐'
    else:
        struct.left_border = '|'
        struct.right_border = '|'
        struct.top_border = '-'
        struct.bottom_border = '-'
        struct.left_bottom_corner = '+'
        struct.right_bottom_corner = '+'
        struct.left_top_corner = '+'
        struct.right_top_corner = '+'


def load_theme(struct, path, colors, default_colors):
    theme = ConfigParser()
    with open(path, 'r') as f:
        theme.readfp(f)
    for k, v in chain(theme.items('syntax'), theme.items('interface')):
        if theme.has_option('syntax', k):
            colors[k] = theme.get('syntax', k)
        else:
            colors[k] = theme.get('interface', k)

    # Check against default theme to see if all values are defined
    for k, v in iteritems(default_colors):
        if k not in colors:
            colors[k] = v



# encoding: utf-8

import re

from bpython.test import unittest
from bpython.line import current_word, current_dict_key, current_dict, \
    current_string, current_object, current_object_attribute, \
    current_from_import_from, current_from_import_import, current_import, \
    current_method_definition_name, current_single_word, \
    current_expression_attribute, current_dotted_attribute


def cursor(s):
    """'ab|c' -> (2, 'abc')"""
    cursor_offset = s.index('|')
    line = s[:cursor_offset] + s[cursor_offset + 1:]
    return cursor_offset, line


def decode(s):
    """'a<bd|c>d' -> ((3, 'abcd'), (1, 3, 'bdc'))"""

    if not s.count('|') == 1:
        raise ValueError('match helper needs | to occur once')
    if s.count('<') != s.count('>') or s.count('<') not in (0, 1):
        raise ValueError('match helper needs <, and > to occur just once')
    matches = list(re.finditer(r'[<>|]', s))
    assert len(matches) in [1, 3], [m.group() for m in matches]
    d = {}
    for i, m in enumerate(matches):
        d[m.group(0)] = m.start() - i
        s = s[:m.start() - i] + s[m.end() - i:]
    assert len(d) in [1, 3], 'need all the parts just once! %r' % d

    if '<' in d:
        return (d['|'], s), (d['<'], d['>'], s[d['<']:d['>']])
    else:
        return (d['|'], s), None


def line_with_cursor(cursor_offset, line):
    return line[:cursor_offset] + '|' + line[cursor_offset:]


def encode(cursor_offset, line, result):
    """encode(3, 'abdcd', (1, 3, 'bdc')) -> a<bd|c>d'

    Written for prettier assert error messages
    """
    encoded_line = line_with_cursor(cursor_offset, line)
    if result is None:
        return encoded_line
    start, end, value = result
    assert line[start:end] == value
    if start < cursor_offset:
        encoded_line = encoded_line[:start] + '<' + encoded_line[start:]
    else:
        encoded_line = (encoded_line[:start + 1] + '<' +
                        encoded_line[start + 1:])
    if end < cursor_offset:
        encoded_line = encoded_line[:end + 1] + '>' + encoded_line[end + 1:]
    else:
        encoded_line = encoded_line[:end + 2] + '>' + encoded_line[end + 2:]
    return encoded_line


class LineTestCase(unittest.TestCase):
    def assertAccess(self, s):
        r"""Asserts that self.func matches as described
        by s, which uses a little language to describe matches:

        abcd<efg>hijklmnopqrstuvwx|yz
           /|\ /|\               /|\
            |   |                 |
         the function should   the current cursor position
         match this "efg"      is between the x and y
        """
        (cursor_offset, line), match = decode(s)
        result = self.func(cursor_offset, line)

        self.assertEqual(
            result, match,
            "%s(%r) result\n%r (%r) doesn't match expected\n%r (%r)" % (
                self.func.__name__, line_with_cursor(cursor_offset, line),
                encode(cursor_offset, line, result), result, s, match))


class TestHelpers(LineTestCase):
    def test_I(self):
        self.assertEqual(cursor('asd|fgh'), (3, 'asdfgh'))

    def test_decode(self):
        self.assertEqual(decode('a<bd|c>d'), ((3, 'abdcd'), (1, 4, 'bdc')))
        self.assertEqual(decode('a|<bdc>d'), ((1, 'abdcd'), (1, 4, 'bdc')))
        self.assertEqual(decode('a<bdc>d|'), ((5, 'abdcd'), (1, 4, 'bdc')))

    def test_encode(self):
        self.assertEqual(encode(3, 'abdcd', (1, 4, 'bdc')), 'a<bd|c>d')
        self.assertEqual(encode(1, 'abdcd', (1, 4, 'bdc')), 'a|<bdc>d')
        self.assertEqual(encode(4, 'abdcd', (1, 4, 'bdc')), 'a<bdc|>d')
        self.assertEqual(encode(5, 'abdcd', (1, 4, 'bdc')), 'a<bdc>d|')

    def test_assert_access(self):
        def dumb_func(cursor_offset, line):
            return (0, 2, 'ab')
        self.func = dumb_func
        self.assertAccess('<a|b>d')


class TestCurrentWord(LineTestCase):
    def setUp(self):
        self.func = current_word

    def test_simple(self):
        self.assertAccess('|')
        self.assertAccess('|asdf')
        self.assertAccess('<a|sdf>')
        self.assertAccess('<asdf|>')
        self.assertAccess('<asdfg|>')
        self.assertAccess('asdf + <asdfg|>')
        self.assertAccess('<asdfg|> + asdf')

    def test_inside(self):
        self.assertAccess('<asd|>')
        self.assertAccess('<asd|fg>')

    def test_dots(self):
        self.assertAccess('<Object.attr1|>')
        self.assertAccess('<Object.attr1.attr2|>')
        self.assertAccess('<Object.att|r1.attr2>')
        self.assertAccess('stuff[stuff] + {123: 456} + <Object.attr1.attr2|>')
        self.assertAccess('stuff[<asd|fg>]')
        self.assertAccess('stuff[asdf[<asd|fg>]')

    def test_non_dots(self):
        self.assertAccess('].asdf|')
        self.assertAccess(').asdf|')
        self.assertAccess('foo[0].asdf|')
        self.assertAccess('foo().asdf|')
        self.assertAccess('foo().|')
        self.assertAccess('foo().asdf.|')
        self.assertAccess('foo[0].asdf.|')

    def test_open_paren(self):
        self.assertAccess('<foo(|>')
        # documenting current behavior - TODO is this intended?


class TestCurrentDictKey(LineTestCase):
    def setUp(self):
        self.func = current_dict_key

    def test_simple(self):
        self.assertAccess('asdf|')
        self.assertAccess('asdf|')
        self.assertAccess('asdf[<>|')
        self.assertAccess('asdf[<>|]')
        self.assertAccess('object.dict[<abc|>')
        self.assertAccess('asdf|')
        self.assertAccess('asdf[<(>|]')
        self.assertAccess('asdf[<(1>|]')
        self.assertAccess('asdf[<(1,>|]')
        self.assertAccess('asdf[<(1, >|]')
        self.assertAccess('asdf[<(1, 2)>|]')
        # TODO self.assertAccess('d[d[<12|>')
        self.assertAccess("d[<'a>|")


class TestCurrentDict(LineTestCase):
    def setUp(self):
        self.func = current_dict

    def test_simple(self):
        self.assertAccess('asdf|')
        self.assertAccess('asdf|')
        self.assertAccess('<asdf>[|')
        self.assertAccess('<asdf>[|]')
        self.assertAccess('<object.dict>[abc|')
        self.assertAccess('asdf|')


class TestCurrentString(LineTestCase):
    def setUp(self):
        self.func = current_string

    def test_closed(self):
        self.assertAccess('"<as|df>"')
        self.assertAccess('"<asdf|>"')
        self.assertAccess('"<|asdf>"')
        self.assertAccess("'<asdf|>'")
        self.assertAccess("'<|asdf>'")
        self.assertAccess("'''<asdf|>'''")
        self.assertAccess('"""<asdf|>"""')
        self.assertAccess('asdf.afd("a") + "<asdf|>"')

    def test_open(self):
        self.assertAccess('"<as|df>')
        self.assertAccess('"<asdf|>')
        self.assertAccess('"<|asdf>')
        self.assertAccess("'<asdf|>")
        self.assertAccess("'<|asdf>")
        self.assertAccess("'''<asdf|>")
        self.assertAccess('"""<asdf|>')
        self.assertAccess('asdf.afd("a") + "<asdf|>')


class TestCurrentObject(LineTestCase):
    def setUp(self):
        self.func = current_object

    def test_simple(self):
        self.assertAccess('<Object>.attr1|')
        self.assertAccess('<Object>.|')
        self.assertAccess('Object|')
        self.assertAccess('Object|.')
        self.assertAccess('<Object>.|')
        self.assertAccess('<Object.attr1>.attr2|')
        self.assertAccess('<Object>.att|r1.attr2')
        self.assertAccess('stuff[stuff] + {123: 456} + <Object.attr1>.attr2|')
        self.assertAccess('stuff[asd|fg]')
        self.assertAccess('stuff[asdf[asd|fg]')


class TestCurrentAttribute(LineTestCase):
    def setUp(self):
        self.func = current_object_attribute

    def test_simple(self):
        self.assertAccess('Object.<attr1|>')
        self.assertAccess('Object.attr1.<attr2|>')
        self.assertAccess('Object.<att|r1>.attr2')
        self.assertAccess('stuff[stuff] + {123: 456} + Object.attr1.<attr2|>')
        self.assertAccess('stuff[asd|fg]')
        self.assertAccess('stuff[asdf[asd|fg]')
        self.assertAccess('Object.attr1.<|attr2>')
        self.assertAccess('Object.<attr1|>.attr2')


class TestCurrentFromImportFrom(LineTestCase):
    def setUp(self):
        self.func = current_from_import_from

    def test_simple(self):
        self.assertAccess('from <sys|> import path')
        self.assertAccess('from <sys> import path|')
        self.assertAccess('if True|: from sys import path')
        self.assertAccess('if True: |from sys import path')
        self.assertAccess('if True: from <sys> import p|ath')
        self.assertAccess('if True: from sys imp|ort path')
        self.assertAccess('if True: from sys import |path')
        self.assertAccess('if True: from sys import path.stu|ff')
        self.assertAccess('if True: from <sys.path> import sep|')
        self.assertAccess('from <os.p|>')


class TestCurrentFromImportImport(LineTestCase):
    def setUp(self):
        self.func = current_from_import_import

    def test_simple(self):
        self.assertAccess('from sys import <path|>')
        self.assertAccess('from sys import <p|ath>')
        self.assertAccess('from sys import |path')
        self.assertAccess('from sys| import path')
        self.assertAccess('from s|ys import path')
        self.assertAccess('from |sys import path')
        self.assertAccess('from xml.dom import <N|ode>')
        # because syntax error
        self.assertAccess('from xml.dom import Node.as|d')


class TestCurrentImport(LineTestCase):
    def setUp(self):
        self.func = current_import

    def test_simple(self):
        self.assertAccess('import <path|>')
        self.assertAccess('import <p|ath>')
        self.assertAccess('import |path')
        self.assertAccess('import path, <another|>')
        self.assertAccess('import path another|')
        self.assertAccess('if True: import <path|>')
        self.assertAccess('if True: import <xml.dom.minidom|>')
        self.assertAccess('if True: import <xml.do|m.minidom>')
        self.assertAccess('if True: import <xml.do|m.minidom> as something')


class TestMethodDefinitionName(LineTestCase):
    def setUp(self):
        self.func = current_method_definition_name

    def test_simple(self):
        self.assertAccess('def <foo|>')
        self.assertAccess('    def bar(x, y)|:')
        self.assertAccess('    def <bar|>(x, y)')


class TestSingleWord(LineTestCase):
    def setUp(self):
        self.func = current_single_word

    def test_simple(self):
        self.assertAccess('foo.bar|')
        self.assertAccess('.foo|')
        self.assertAccess(' <foo|>')


class TestCurrentExpressionAttribute(LineTestCase):
    def setUp(self):
        self.func = current_expression_attribute

    def test_simple(self):
        self.assertAccess('Object.<attr1|>.')
        self.assertAccess('Object.<|attr1>.')
        self.assertAccess('Object.(|)')
        self.assertAccess('Object.another.(|)')
        self.assertAccess('asdf asdf asdf.(abc|)')

    def test_without_dot(self):
        self.assertAccess('Object|')
        self.assertAccess('Object|.')
        self.assertAccess('|Object.')

    def test_with_whitespace(self):
        self.assertAccess('Object. <attr|>')
        self.assertAccess('Object .<attr|>')
        self.assertAccess('Object . <attr|>')
        self.assertAccess('Object .asdf attr|')
        self.assertAccess('Object .<asdf|> attr')
        self.assertAccess('Object. asdf attr|')
        self.assertAccess('Object. <asdf|> attr')
        self.assertAccess('Object . asdf attr|')
        self.assertAccess('Object . <asdf|> attr')

    def test_indexing(self):
        self.assertAccess('abc[def].<ghi|>')
        self.assertAccess('abc[def].<|ghi>')
        self.assertAccess('abc[def].<gh|i>')
        self.assertAccess('abc[def].gh |i')
        self.assertAccess('abc[def]|')

    def test_strings(self):
        self.assertAccess('"hey".<a|>')
        self.assertAccess('"hey"|')
        self.assertAccess('"hey"|.a')
        self.assertAccess('"hey".<a|b>')
        self.assertAccess('"hey".asdf d|')
        self.assertAccess('"hey".<|>')


class TestCurrentDottedAttribute(LineTestCase):
    def setUp(self):
        self.func = current_dotted_attribute

    def test_simple(self):
        self.assertAccess('<obj.attr>|')
        self.assertAccess('(<obj.attr>|')
        self.assertAccess('[<obj.attr>|')
        self.assertAccess('m.body[0].value|')
        self.assertAccess('m.body[0].attr.value|')

if __name__ == '__main__':
    unittest.main()



# encoding: utf-8

"""implementations of simple readline edit operations

just the ones that fit the model of transforming the current line
and the cursor location
based on http://www.bigsmoke.us/readline/shortcuts"""

from bpython.lazyre import LazyReCompile

import inspect
from six import iteritems
from bpython._py3compat import py3

INDENT = 4

# TODO Allow user config of keybindings for these actions
if not py3:
    getargspec = lambda func: inspect.getargspec(func)[0]
else:
    getargspec = lambda func: inspect.signature(func).parameters


class AbstractEdits(object):

    default_kwargs = {
        'line': 'hello world',
        'cursor_offset': 5,
        'cut_buffer': 'there',
    }

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def add(self, key, func, overwrite=False):
        if key in self:
            if overwrite:
                del self[key]
            else:
                raise ValueError('key %r already has a mapping' % (key,))
        params = getargspec(func)
        args = dict((k, v) for k, v in iteritems(self.default_kwargs)
                    if k in params)
        r = func(**args)
        if len(r) == 2:
            if hasattr(func, 'kills'):
                raise ValueError('function %r returns two values, but has a '
                                 'kills attribute' % (func,))
            self.simple_edits[key] = func
        elif len(r) == 3:
            if not hasattr(func, 'kills'):
                raise ValueError('function %r returns three values, but has '
                                 'no kills attribute' % (func,))
            self.cut_buffer_edits[key] = func
        else:
            raise ValueError('return type of function %r not recognized' %
                             (func,))

    def add_config_attr(self, config_attr, func):
        if config_attr in self.awaiting_config:
            raise ValueError('config attribute %r already has a mapping' %
                             (config_attr,))
        self.awaiting_config[config_attr] = func

    def call(self, key, **kwargs):
        func = self[key]
        params = getargspec(func)
        args = dict((k, v) for k, v in kwargs.items() if k in params)
        return func(**args)

    def call_without_cut(self, key, **kwargs):
        """Looks up the function and calls it, returning only line and cursor
        offset"""
        r = self.call_for_two(key, **kwargs)
        return r[:2]

    def __getitem__(self, key):
        if key in self.simple_edits:
            return self.simple_edits[key]
        if key in self.cut_buffer_edits:
            return self.cut_buffer_edits[key]
        raise KeyError("key %r not mapped" % (key,))

    def __delitem__(self, key):
        if key in self.simple_edits:
            del self.simple_edits[key]
        elif key in self.cut_buffer_edits:
            del self.cut_buffer_edits[key]
        else:
            raise KeyError("key %r not mapped" % (key,))


class UnconfiguredEdits(AbstractEdits):
    """Maps key to edit functions, and bins them by what parameters they take.

    Only functions with specific signatures can be added:
        * func(**kwargs) -> cursor_offset, line
        * func(**kwargs) -> cursor_offset, line, cut_buffer
        where kwargs are in among the keys of Edits.default_kwargs
    These functions will be run to determine their return type, so no side
    effects!

    More concrete Edits instances can be created by applying a config with
    Edits.mapping_with_config() - this creates a new Edits instance
    that uses a config file to assign config_attr bindings.

    Keys can't be added twice, config attributes can't be added twice.
    """

    def __init__(self):
        self.simple_edits = {}
        self.cut_buffer_edits = {}
        self.awaiting_config = {}

    def mapping_with_config(self, config, key_dispatch):
        """Creates a new mapping object by applying a config object"""
        return ConfiguredEdits(self.simple_edits, self.cut_buffer_edits,
                               self.awaiting_config, config, key_dispatch)

    def on(self, key=None, config=None):
        if not ((key is None) ^ (config is None)):
            raise ValueError("Must use exactly one of key, config")
        if key is not None:
            def add_to_keybinds(func):
                self.add(key, func)
                return func
            return add_to_keybinds
        else:
            def add_to_config(func):
                self.add_config_attr(config, func)
                return func
            return add_to_config


class ConfiguredEdits(AbstractEdits):
    def __init__(self, simple_edits, cut_buffer_edits, awaiting_config, config,
                 key_dispatch):
        self.simple_edits = dict(simple_edits)
        self.cut_buffer_edits = dict(cut_buffer_edits)
        for attr, func in awaiting_config.items():
            for key in key_dispatch[getattr(config, attr)]:
                super(ConfiguredEdits, self).add(key, func, overwrite=True)

    def add_config_attr(self, config_attr, func):
        raise NotImplementedError("Config already set on this mapping")

    def add(self, key, func):
        raise NotImplementedError("Config already set on this mapping")


edit_keys = UnconfiguredEdits()

# Because the edits.on decorator runs the functions, functions which depend
# on other functions must be declared after their dependencies


def kills_behind(func):
    func.kills = 'behind'
    return func


def kills_ahead(func):
    func.kills = 'ahead'
    return func


@edit_keys.on(config='left_key')
@edit_keys.on('<LEFT>')
def left_arrow(cursor_offset, line):
    return max(0, cursor_offset - 1), line


@edit_keys.on(config='right_key')
@edit_keys.on('<RIGHT>')
def right_arrow(cursor_offset, line):
    return min(len(line), cursor_offset + 1), line


@edit_keys.on(config='beginning_of_line_key')
@edit_keys.on('<HOME>')
def beginning_of_line(cursor_offset, line):
    return 0, line


@edit_keys.on(config='end_of_line_key')
@edit_keys.on('<END>')
def end_of_line(cursor_offset, line):
    return len(line), line


forward_word_re = LazyReCompile(r"\S\s")


@edit_keys.on('<Esc+f>')
@edit_keys.on('<Ctrl-RIGHT>')
@edit_keys.on('<Esc+RIGHT>')
def forward_word(cursor_offset, line):
    match = forward_word_re.search(line[cursor_offset:]+' ')
    delta = match.end() - 1 if match else 0
    return (cursor_offset + delta, line)


def last_word_pos(string):
    """returns the start index of the last word of given string"""
    match = forward_word_re.search(string[::-1])
    index = match and len(string) - match.end() + 1
    return index or 0


@edit_keys.on('<Esc+b>')
@edit_keys.on('<Ctrl-LEFT>')
@edit_keys.on('<Esc+LEFT>')
def back_word(cursor_offset, line):
    return (last_word_pos(line[:cursor_offset]), line)


@edit_keys.on('<DELETE>')
def delete(cursor_offset, line):
    return (cursor_offset,
            line[:cursor_offset] + line[cursor_offset+1:])


@edit_keys.on('<BACKSPACE>')
@edit_keys.on(config='backspace_key')
def backspace(cursor_offset, line):
    if cursor_offset == 0:
        return cursor_offset, line
    if not line[:cursor_offset].strip():  # if just whitespace left of cursor
        # front_white = len(line[:cursor_offset]) - \
        #     len(line[:cursor_offset].lstrip())
        to_delete = ((cursor_offset - 1) % INDENT) + 1
        return (cursor_offset - to_delete,
                line[:cursor_offset - to_delete] + line[cursor_offset:])
    return (cursor_offset - 1,
            line[:cursor_offset - 1] + line[cursor_offset:])


@edit_keys.on(config='clear_line_key')
def delete_from_cursor_back(cursor_offset, line):
    return 0, line[cursor_offset:]


delete_rest_of_word_re = LazyReCompile(r'\w\b')


@edit_keys.on('<Esc+d>')  # option-d
@kills_ahead
def delete_rest_of_word(cursor_offset, line):
    m = delete_rest_of_word_re.search(line[cursor_offset:])
    if not m:
        return cursor_offset, line, ''
    return (cursor_offset,
            line[:cursor_offset] + line[m.start()+cursor_offset+1:],
            line[cursor_offset:m.start()+cursor_offset+1])


delete_word_to_cursor_re = LazyReCompile(r'\s\S')


@edit_keys.on(config='clear_word_key')
@kills_behind
def delete_word_to_cursor(cursor_offset, line):
    start = 0
    for match in delete_word_to_cursor_re.finditer(line[:cursor_offset]):
        start = match.start() + 1
    return (start, line[:start] + line[cursor_offset:],
            line[start:cursor_offset])


@edit_keys.on('<Esc+y>')
def yank_prev_prev_killed_text(cursor_offset, line, cut_buffer):
    # TODO not implemented - just prev
    return (cursor_offset+len(cut_buffer),
            line[:cursor_offset] + cut_buffer + line[cursor_offset:])


@edit_keys.on(config='yank_from_buffer_key')
def yank_prev_killed_text(cursor_offset, line, cut_buffer):
    return (cursor_offset+len(cut_buffer),
            line[:cursor_offset] + cut_buffer + line[cursor_offset:])


@edit_keys.on(config='transpose_chars_key')
def transpose_character_before_cursor(cursor_offset, line):
    if cursor_offset < 2:
        return cursor_offset, line
    if cursor_offset == len(line):
        return cursor_offset, line[:-2] + line[-1] + line[-2]
    return (min(len(line), cursor_offset + 1),
            line[:cursor_offset - 1] +
            (line[cursor_offset] if len(line) > cursor_offset else '') +
            line[cursor_offset - 1] +
            line[cursor_offset + 1:])


@edit_keys.on('<Esc+t>')
def transpose_word_before_cursor(cursor_offset, line):
    return cursor_offset, line  # TODO Not implemented

# TODO undo all changes to line: meta-r

# bonus functions (not part of readline)


@edit_keys.on('<Esc+u>')
def uppercase_next_word(cursor_offset, line):
    return cursor_offset, line  # TODO Not implemented


@edit_keys.on(config='cut_to_buffer_key')
@kills_ahead
def delete_from_cursor_forward(cursor_offset, line):
    return cursor_offset, line[:cursor_offset], line[cursor_offset:]


@edit_keys.on('<Esc+c>')
def titlecase_next_word(cursor_offset, line):
    return cursor_offset, line  # TODO Not implemented


delete_word_from_cursor_back_re = LazyReCompile(r'^|\b\w')


@edit_keys.on('<Esc+BACKSPACE>')
@edit_keys.on('<Meta-BACKSPACE>')
@kills_behind
def delete_word_from_cursor_back(cursor_offset, line):
    """Whatever my option-delete does in bash on my mac"""
    if not line:
        return cursor_offset, line, ''
    start = None
    for match in delete_word_from_cursor_back_re.finditer(line):
        if match.start() < cursor_offset:
            start = match.start()
    if start is not None:
        return (start, line[:start] + line[cursor_offset:],
                line[start:cursor_offset])
    else:
        return cursor_offset, line, ''



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import sys
import re
from textwrap import dedent

from curtsies.fmtfuncs import bold, green, magenta, cyan, red, plain

from bpython.curtsiesfrontend import interpreter
from bpython._py3compat import py3
from bpython.test import mock, unittest

pypy = 'PyPy' in sys.version


def remove_ansi(s):
    return re.sub(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]'.encode('ascii'), b'', s)


class TestInterpreter(unittest.TestCase):
    def interp_errlog(self):
        i = interpreter.Interp()
        a = []
        i.write = a.append
        return i, a

    def err_lineno(self, a):
        strings = [x.__unicode__() for x in a]
        for line in reversed(strings):
            clean_line = remove_ansi(line)
            m = re.search(r'line (\d+)[,]', clean_line)
            if m:
                return int(m.group(1))
        return None

    def test_syntaxerror(self):
        i, a = self.interp_errlog()

        i.runsource('1.1.1.1')

        if pypy:
            expected = (
                '  File ' + green('"<input>"') +
                ', line ' + bold(magenta('1')) + '\n    1.1.1.1\n      ^\n' +
                bold(red('SyntaxError')) + ': ' + cyan('invalid syntax') +
                '\n')
        else:
            expected = (
                '  File ' + green('"<input>"') +
                ', line ' + bold(magenta('1')) + '\n    1.1.1.1\n        ^\n' +
                bold(red('SyntaxError')) + ': ' + cyan('invalid syntax') +
                '\n')

        self.assertMultiLineEqual(str(plain('').join(a)), str(expected))
        self.assertEqual(plain('').join(a), expected)

    def test_traceback(self):
        i, a = self.interp_errlog()

        def f():
            return 1 / 0

        def gfunc():
            return f()

        i.runsource('gfunc()')

        if pypy and not py3:
            global_not_found = "global name 'gfunc' is not defined"
        else:
            global_not_found = "name 'gfunc' is not defined"

        expected = (
            'Traceback (most recent call last):\n  File ' +
            green('"<input>"') + ', line ' +
            bold(magenta('1')) + ', in ' + cyan('<module>') + '\n    gfunc()\n' +
            bold(red('NameError')) + ': ' + cyan(global_not_found) + '\n')

        self.assertMultiLineEqual(str(plain('').join(a)), str(expected))
        self.assertEqual(plain('').join(a), expected)

    @unittest.skipIf(py3, "runsource() accepts only unicode in Python 3")
    def test_runsource_bytes(self):
        i = interpreter.Interp(encoding=b'latin-1')

        i.runsource("a = b'\xfe'".encode('latin-1'), encode=False)
        self.assertIsInstance(i.locals['a'], str)
        self.assertEqual(i.locals['a'], b"\xfe")

        i.runsource("b = u'\xfe'".encode('latin-1'), encode=False)
        self.assertIsInstance(i.locals['b'], unicode)
        self.assertEqual(i.locals['b'], "\xfe")

    @unittest.skipUnless(py3, "Only a syntax error in Python 3")
    def test_runsource_bytes_over_128_syntax_error_py3(self):
        i = interpreter.Interp(encoding=b'latin-1')
        i.showsyntaxerror = mock.Mock(return_value=None)

        i.runsource("a = b'\xfe'")
        i.showsyntaxerror.assert_called_with(mock.ANY)

    @unittest.skipIf(py3, "encode is Python 2 only")
    def test_runsource_bytes_over_128_syntax_error_py2(self):
        i = interpreter.Interp(encoding=b'latin-1')

        i.runsource(b"a = b'\xfe'")
        self.assertIsInstance(i.locals['a'], type(b''))
        self.assertEqual(i.locals['a'], b"\xfe")

    @unittest.skipIf(py3, "encode is Python 2 only")
    def test_runsource_unicode(self):
        i = interpreter.Interp(encoding=b'latin-1')

        i.runsource("a = u'\xfe'")
        self.assertIsInstance(i.locals['a'], type(u''))
        self.assertEqual(i.locals['a'], u"\xfe")

    def test_getsource_works_on_interactively_defined_functions(self):
        source = 'def foo(x):\n    return x + 1\n'
        i = interpreter.Interp()
        i.runsource(source)
        import inspect
        inspected_source = inspect.getsource(i.locals['foo'])
        self.assertEqual(inspected_source, source)

    @unittest.skipIf(py3, "encode only does anything in Python 2")
    def test_runsource_unicode_autoencode_and_noencode(self):
        """error line numbers should be fixed"""

        # Since correct behavior for unicode is the same
        # for auto and False, run the same tests
        for encode in ['auto', False]:
            i, a = self.interp_errlog()
            i.runsource(u'[1 + 1,\nabcd]', encode=encode)
            self.assertEqual(self.err_lineno(a), 2)

            i, a = self.interp_errlog()
            i.runsource(u'[1 + 1,\nabcd]', encode=encode)
            self.assertEqual(self.err_lineno(a), 2)

            i, a = self.interp_errlog()
            i.runsource(u'#encoding: utf-8\nabcd', encode=encode)
            self.assertEqual(self.err_lineno(a), 2)

            i, a = self.interp_errlog()
            i.runsource(u'#encoding: utf-8\nabcd',
                        filename='x.py', encode=encode)
            self.assertIn('SyntaxError:',
                          ''.join(''.join(remove_ansi(x.__unicode__())
                                          for x in a)))

    @unittest.skipIf(py3, "encode only does anything in Python 2")
    def test_runsource_unicode_encode(self):
        i, _ = self.interp_errlog()
        with self.assertRaises(ValueError):
            i.runsource(u'1 + 1', encode=True)

        i, _ = self.interp_errlog()
        with self.assertRaises(ValueError):
            i.runsource(u'1 + 1', filename='x.py', encode=True)

    @unittest.skipIf(py3, "encode only does anything in Python 2")
    def test_runsource_bytestring_noencode(self):
        i, a = self.interp_errlog()
        i.runsource(b'[1 + 1,\nabcd]', encode=False)
        self.assertEqual(self.err_lineno(a), 2)

        i, a = self.interp_errlog()
        i.runsource(b'[1 + 1,\nabcd]', filename='x.py', encode=False)
        self.assertEqual(self.err_lineno(a), 2)

        i, a = self.interp_errlog()
        i.runsource(dedent(b'''\
                    #encoding: utf-8

                    ["%s",
                    abcd]''' % (u'åß∂ƒ'.encode('utf8'),)), encode=False)
        self.assertEqual(self.err_lineno(a), 4)

        i, a = self.interp_errlog()
        i.runsource(dedent(b'''\
                    #encoding: utf-8

                    ["%s",
                    abcd]''' % (u'åß∂ƒ'.encode('utf8'),)),
                    filename='x.py', encode=False)
        self.assertEqual(self.err_lineno(a), 4)

    @unittest.skipIf(py3, "encode only does anything in Python 2")
    def test_runsource_bytestring_encode(self):
        i, a = self.interp_errlog()
        i.runsource(b'[1 + 1,\nabcd]', encode=True)
        self.assertEqual(self.err_lineno(a), 2)

        i, a = self.interp_errlog()
        with self.assertRaises(ValueError):
            i.runsource(b'[1 + 1,\nabcd]', filename='x.py', encode=True)

        i, a = self.interp_errlog()
        i.runsource(dedent(b'''\
                    #encoding: utf-8

                    [u"%s",
                    abcd]''' % (u'åß∂ƒ'.encode('utf8'),)), encode=True)
        self.assertEqual(self.err_lineno(a), 4)

        i, a = self.interp_errlog()
        with self.assertRaises(ValueError):
            i.runsource(dedent(b'''\
                        #encoding: utf-8

                        [u"%s",
                        abcd]''' % (u'åß∂ƒ'.encode('utf8'),)),
                        filename='x.py',
                        encode=True)

    @unittest.skipIf(py3, "encode only does anything in Python 2")
    def test_runsource_bytestring_autoencode(self):
        i, a = self.interp_errlog()
        i.runsource(b'[1 + 1,\n abcd]')
        self.assertEqual(self.err_lineno(a), 2)

        i, a = self.interp_errlog()
        i.runsource(b'[1 + 1,\nabcd]', filename='x.py')
        self.assertEqual(self.err_lineno(a), 2)

        i, a = self.interp_errlog()
        i.runsource(dedent(b'''\
                    #encoding: utf-8

                    [u"%s",
                    abcd]''' % (u'åß∂ƒ'.encode('utf8'),)))
        self.assertEqual(self.err_lineno(a), 4)

        i, a = self.interp_errlog()
        i.runsource(dedent(b'''\
                    #encoding: utf-8

                    [u"%s",
                    abcd]''' % (u'åß∂ƒ'.encode('utf8'),)))
        self.assertEqual(self.err_lineno(a), 4)



# encoding: utf-8

# The MIT License
#
# Copyright (c) 2008 Simon de Vlieger
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import string
from six.moves import range


class KeyMap(object):

    def __init__(self, default=''):
        self.map = {}
        self.default = default

    def __getitem__(self, key):
        if not key:
            # Unbound key
            return self.default
        elif key in self.map:
            return self.map[key]
        else:
            raise KeyError('Configured keymap (%s)' % key +
                           ' does not exist in bpython.keys')

    def __delitem__(self, key):
        del self.map[key]

    def __setitem__(self, key, value):
        self.map[key] = value


cli_key_dispatch = KeyMap(tuple())
urwid_key_dispatch = KeyMap('')

# fill dispatch with letters
for c in string.ascii_lowercase:
    cli_key_dispatch['C-%s' % c] = (chr(string.ascii_lowercase.index(c) + 1),
                                    '^%s' % c.upper())

for c in string.ascii_lowercase:
    urwid_key_dispatch['C-%s' % c] = 'ctrl %s' % c
    urwid_key_dispatch['M-%s' % c] = 'meta %s' % c

# fill dispatch with cool characters
cli_key_dispatch['C-['] = (chr(27), '^[')
cli_key_dispatch['C-\\'] = (chr(28), '^\\')
cli_key_dispatch['C-]'] = (chr(29), '^]')
cli_key_dispatch['C-^'] = (chr(30), '^^')
cli_key_dispatch['C-_'] = (chr(31), '^_')

# fill dispatch with function keys
for x in range(1, 13):
    cli_key_dispatch['F%d' % x] = ('KEY_F(%d)' % x,)

for x in range(1, 13):
    urwid_key_dispatch['F%d' % x] = 'f%d' % x



# coding: utf-8
from __future__ import unicode_literals

from functools import partial
import re

from bpython.lazyre import LazyReCompile

from curtsies.termformatconstants import FG_COLORS, BG_COLORS, colors
from curtsies.formatstring import fmtstr, FmtStr


cnames = dict(zip('krgybmcwd', colors + ('default',)))


def func_for_letter(l, default='k'):
    """Returns FmtStr constructor for a bpython-style color code"""
    if l == 'd':
        l = default
    elif l == 'D':
        l = default.upper()
    return partial(fmtstr, fg=cnames[l.lower()], bold=l.isupper())


def color_for_letter(l, default='k'):
    if l == 'd':
        l = default
    return cnames[l.lower()]


def parse(s):
    """Returns a FmtStr object from a bpython-formatted colored string"""
    rest = s
    stuff = []
    while True:
        if not rest:
            break
        start, rest = peel_off_string(rest)
        stuff.append(start)
    return (sum((fs_from_match(d) for d in stuff[1:]), fs_from_match(stuff[0]))
            if len(stuff) > 0
            else FmtStr())


def fs_from_match(d):
    atts = {}
    if d['fg']:

        # this isn't according to spec as I understand it
        if d['fg'].isupper():
            d['bold'] = True
        # TODO figure out why boldness isn't based on presence of \x02

        color = cnames[d['fg'].lower()]
        if color != 'default':
            atts['fg'] = FG_COLORS[color]
    if d['bg']:
        if d['bg'] == 'I':
            # hack for finding the "inverse"
            color = colors[(colors.index(color) + (len(colors) // 2)) %
                           len(colors)]
        else:
            color = cnames[d['bg'].lower()]
        if color != 'default':
            atts['bg'] = BG_COLORS[color]
    if d['bold']:
        atts['bold'] = True
    return fmtstr(d['string'], **atts)


peel_off_string_re = LazyReCompile(
    r"""(?P<colormarker>\x01
            (?P<fg>[krgybmcwdKRGYBMCWD]?)
            (?P<bg>[krgybmcwdKRGYBMCWDI]?)?)
        (?P<bold>\x02?)
        \x03
        (?P<string>[^\x04]*)
        \x04
        (?P<rest>.*)
        """, re.VERBOSE | re.DOTALL)


def peel_off_string(s):
    m = peel_off_string_re.match(s)
    assert m, repr(s)
    d = m.groupdict()
    rest = d['rest']
    del d['rest']
    return d, rest



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import itertools
from six.moves import range

from curtsies import fsarray, fmtstr, FSArray
from curtsies.formatstring import linesplit
from curtsies.fmtfuncs import bold

from bpython.curtsiesfrontend.parse import func_for_letter
from bpython._py3compat import py3
if not py3:
    import inspect

logger = logging.getLogger(__name__)

# All paint functions should
# * return an array of the width they were asked for
# * return an array not taller than the height they were asked for


def display_linize(msg, columns, blank_line=False):
    """Returns lines obtained by splitting msg over multiple lines.

    Warning: if msg is empty, returns an empty list of lines"""
    display_lines = ([msg[start:end]
                      for start, end in zip(
                          range(0, len(msg), columns),
                          range(columns, len(msg) + columns, columns))]
                     if msg else ([''] if blank_line else []))
    return display_lines


def paint_history(rows, columns, display_lines):
    lines = []
    for r, line in zip(range(rows), display_lines[-rows:]):
        lines.append(fmtstr(line[:columns]))
    r = fsarray(lines, width=columns)
    assert r.shape[0] <= rows, repr(r.shape) + ' ' + repr(rows)
    assert r.shape[1] <= columns, repr(r.shape) + ' ' + repr(columns)
    return r


def paint_current_line(rows, columns, current_display_line):
    lines = display_linize(current_display_line, columns, True)
    return fsarray(lines, width=columns)


def paginate(rows, matches, current, words_wide):
    if current not in matches:
        current = matches[0]
    per_page = rows * words_wide
    current_page = matches.index(current) // per_page
    return matches[per_page * current_page:per_page * (current_page + 1)]


def matches_lines(rows, columns, matches, current, config, match_format):
    highlight_color = func_for_letter(config.color_scheme['operator'].lower())

    if not matches:
        return []
    color = func_for_letter(config.color_scheme['main'])
    max_match_width = max(len(m) for m in matches)
    words_wide = max(1, (columns - 1) // (max_match_width + 1))
    matches = [match_format(m) for m in matches]
    if current:
        current = match_format(current)

    matches = paginate(rows, matches, current, words_wide)

    result = [fmtstr(' ').join(color(m.ljust(max_match_width))
                                     if m != current
                                     else highlight_color(
                                         m.ljust(max_match_width))
                                     for m in matches[i:i + words_wide])
                     for i in range(0, len(matches), words_wide)]

    logger.debug('match: %r' % current)
    logger.debug('matches_lines: %r' % result)
    return result


def formatted_argspec(funcprops, arg_pos, columns, config):
    # Pretty directly taken from bpython.cli
    func = funcprops.func
    args = funcprops.argspec.args
    kwargs = funcprops.argspec.defaults
    _args = funcprops.argspec.varargs
    _kwargs = funcprops.argspec.varkwargs
    is_bound_method = funcprops.is_bound_method
    if py3:
        kwonly = funcprops.argspec.kwonly
        kwonly_defaults = funcprops.argspec.kwonly_defaults or dict()

    arg_color = func_for_letter(config.color_scheme['name'])
    func_color = func_for_letter(config.color_scheme['name'].swapcase())
    punctuation_color = func_for_letter(config.color_scheme['punctuation'])
    token_color = func_for_letter(config.color_scheme['token'])
    bolds = {token_color: lambda x: bold(token_color(x)),
             arg_color: lambda x: bold(arg_color(x))}

    s = func_color(func) + arg_color(': (')

    if is_bound_method and isinstance(arg_pos, int):
        # TODO what values could this have?
        arg_pos += 1

    for i, arg in enumerate(args):
        kw = None
        if kwargs and i >= len(args) - len(kwargs):
            kw = str(kwargs[i - (len(args) - len(kwargs))])
        color = token_color if arg_pos in (i, arg) else arg_color
        if i == arg_pos or arg == arg_pos:
            color = bolds[color]

        if not py3:
            s += color(inspect.strseq(arg, unicode))
        else:
            s += color(arg)

        if kw is not None:
            s += punctuation_color('=')
            if not py3:
                kw = kw.decode('ascii', 'replace')
            s += token_color(kw)

        if i != len(args) - 1:
            s += punctuation_color(', ')

    if _args:
        if args:
            s += punctuation_color(', ')
        s += token_color('*%s' % (_args,))

    if py3 and kwonly:
        if not _args:
            if args:
                s += punctuation_color(', ')
            s += punctuation_color('*')
        marker = object()
        for arg in kwonly:
            s += punctuation_color(', ')
            color = token_color
            if arg_pos:
                color = bolds[color]
            s += color(arg)
            default = kwonly_defaults.get(arg, marker)
            if default is not marker:
                s += punctuation_color('=')
                s += token_color(repr(default))

    if _kwargs:
        if args or _args or (py3 and kwonly):
            s += punctuation_color(', ')
        s += token_color('**%s' % (_kwargs,))
    s += punctuation_color(')')

    return linesplit(s, columns)


def formatted_docstring(docstring, columns, config):
    if isinstance(docstring, bytes):
        docstring = docstring.decode('utf8')
    elif isinstance(docstring, str if py3 else unicode):
        pass
    else:
        # TODO: fail properly here and catch possible exceptions in callers.
        return []
    color = func_for_letter(config.color_scheme['comment'])
    return sum(([color(x) for x in (display_linize(line, columns) if line else
                                    fmtstr(''))]
                for line in docstring.split('\n')), [])


def paint_infobox(rows, columns, matches, funcprops, arg_pos, match, docstring,
                  config, match_format):
    """Returns painted completions, funcprops, match, docstring etc."""
    if not (rows and columns):
        return FSArray(0, 0)
    width = columns - 4
    from_argspec = (formatted_argspec(funcprops, arg_pos, width, config)
                    if funcprops else [])
    from_doc = (formatted_docstring(docstring, width, config)
                if docstring else [])
    from_matches = (matches_lines(max(1, rows - len(from_argspec) - 2),
                                  width, matches, match, config, match_format)
                    if matches else [])

    lines = from_argspec + from_matches + from_doc

    def add_border(line):
        """Add colored borders left and right to a line."""
        new_line = border_color(config.left_border + ' ')
        new_line += line.ljust(width)[:width]
        new_line += border_color(' ' + config.right_border)
        return new_line

    border_color = func_for_letter(config.color_scheme['main'])

    top_line = border_color(config.left_top_corner +
                            config.top_border * (width + 2) +
                            config.right_top_corner)
    bottom_line = border_color(config.left_bottom_corner +
                               config.bottom_border * (width + 2) +
                               config.right_bottom_corner)

    output_lines = list(itertools.chain((top_line, ), map(add_border, lines),
                                        (bottom_line, )))
    r = fsarray(output_lines[:min(rows - 1,
                                  len(output_lines) - 1)] + output_lines[-1:])
    return r


def paint_last_events(rows, columns, names, config):
    if not names:
        return fsarray([])
    width = min(max(len(name) for name in names), columns - 2)
    output_lines = []
    output_lines.append(config.left_top_corner + config.top_border * width +
                        config.right_top_corner)
    for name in reversed(names[max(0, len(names) - (rows - 2)):]):
        output_lines.append(config.left_border + name[:width].center(width) +
                            config.right_border)
    output_lines.append(config.left_bottom_corner +
                        config.bottom_border * width +
                        config.right_bottom_corner)
    return fsarray(output_lines)


def paint_statusbar(rows, columns, msg, config):
    func = func_for_letter(config.color_scheme['main'])
    return fsarray([func(msg.ljust(columns))[:columns]])
